from collections import defaultdict
from typing import Iterable, TypeVar, Dict, List, Callable, Hashable

T = TypeVar("T")
K = TypeVar("K", bound=Hashable, covariant=True)


def group_by(collection: Iterable[T], key: Callable[[T], K]) -> Dict[K, List[T]]:
    res: Dict[K, List[T]] = defaultdict(list)
    for value in collection:
        res[key(value)].append(value)
    return res



import asyncio
from os import environ
from typing import Dict


async def run_os_cmd(cmd: str, env: Dict[str, str] = {}, echo: bool = True) -> str:
    """Run command in shell return output and print stdout in real time with the echo parameter

    Args:
        cmd (str): Shell command
        env (Dict[str, str], optional): Environment variables for the process. Defaults to {}.
        echo (bool, optional): Should print stdout in real time. Defaults to True.

    Raises:
        RuntimeError: On !=0 returncode throw exception with stdout and stderr

    Returns:
        str: stdout as string decoded
    """
    msg = f"runnning cmd: '{cmd}'"
    if env:
        envs_str = ";".join(f'{k}="{v}"' for k, v in env.items())
        msg += f" with {envs_str}"
    print(msg)

    _env = environ.copy()
    _env.update(env)

    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        env=_env,
    )

    output_str = ""
    communicate_cor = proc.communicate()

    assert proc.stdout is not None
    async for chunk in proc.stdout:
        chunk_decoded = chunk.decode()
        if echo:
            print(chunk_decoded, end="")
        output_str += chunk_decoded

    stdout, stderr = await communicate_cor
    if proc.returncode != 0:
        msgs = (
            f"Command '{cmd}' exited with {proc.returncode}",
            "stdout:",
            output_str + stdout.decode(),
            "stderr:",
            stderr.decode(),
        )
        raise RuntimeError("\n".join(msgs))
    return output_str



import pytest
from io import StringIO
import logging
from clubbi_utils.json_logging import JsonLogger
from pydantic.dataclasses import dataclass
from pydantic import BaseModel

class ExampleBaseModel(BaseModel):
    n: int
    s: str

@dataclass
class ExamplePydanticDataclass:
    n: int
    s: str


@pytest.fixture
def mocked_stdout() -> StringIO:
    return StringIO()


@pytest.fixture
def mocked_logging(mocked_stdout: StringIO) -> logging.Logger:
    logger = logging.getLogger()
    logger.addHandler
    logger.addHandler(logging.StreamHandler(mocked_stdout))
    return logger


@pytest.fixture
def jlogger(mocked_logging: logging.Logger) -> JsonLogger:
    return JsonLogger(mocked_logging)


def test_json_logging(jlogger: JsonLogger, mocked_stdout: StringIO) -> None:
    jlogger.info("main", "test", input=1)
    assert mocked_stdout.getvalue().strip() == r'{"workflow":"main","message":"test","level":"INFO","input":1}'

def test_json_logging_with_base_model(jlogger: JsonLogger, mocked_stdout: StringIO) -> None:
    jlogger.info("main", "test", base_model=ExampleBaseModel(n=42, s="hello world"))
    assert mocked_stdout.getvalue().strip() == r'{"workflow":"main","message":"test","level":"INFO","base_model":{"n":42,"s":"hello world"}}'

def test_json_logging_with_pydantic_dataclass(jlogger: JsonLogger, mocked_stdout: StringIO) -> None:
    jlogger.info("main", "test", dataclass=ExamplePydanticDataclass(n=42, s="hello world"))
    assert mocked_stdout.getvalue().strip() == r'{"workflow":"main","message":"test","level":"INFO","dataclass":{"n":42,"s":"hello world"}}'

def test_json_logging_with_dict(jlogger: JsonLogger, mocked_stdout: StringIO) -> None:
    jlogger.info("main", "test", dict=dict(n=42, s="hello world"))
    assert mocked_stdout.getvalue().strip() == r'{"workflow":"main","message":"test","level":"INFO","dict":{"n":42,"s":"hello world"}}'


from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from os import getenv

from aiobotocore.client import AioBaseClient

from clubbi_utils.logging import logger
from clubbi_utils.mailer import Email

CLUBBI_CONTROL_EMAIL = getenv("CLUBBI_CONTROL_EMAIL", "success@simulator.amazonses.com")


class SesMailer:
    def __init__(self, client: AioBaseClient):
        self.client = client
        self._charset = "UTF-8"

    async def send(self, email: Email) -> None:
        if email.attachments:
            response = await self.client.send_raw_email(**self._build_raw_email(email))
        else:
            response = await self.client.send_email(**self._build_email_data(email))
        logger.info(response["MessageId"])

    def _build_email_data(self, email: Email) -> dict:
        return dict(
            Destination={"ToAddresses": email.recipients, "BccAddresses": [CLUBBI_CONTROL_EMAIL]},
            Message={
                "Body": {
                    "Html": {
                        "Charset": self._charset,
                        "Data": email.html_body,
                    },
                    "Text": {
                        "Charset": self._charset,
                        "Data": email.raw_text_body,
                    },
                },
                "Subject": {
                    "Charset": self._charset,
                    "Data": email.subject,
                },
            },
            Source=email.sender,
        )

    def _build_raw_email(self, email: Email) -> dict:
        msg = MIMEMultipart()
        msg["Subject"] = email.subject
        msg["From"] = email.sender
        msg["To"] = email.recipients[0]

        msg_body = MIMEMultipart("alternative")
        # the message body
        text_part = MIMEText(email.raw_text_body, "plain", self._charset)
        html_part = MIMEText(email.html_body, "html", self._charset)

        # Add the text and HTML parts to the child container.
        msg_body.attach(text_part)
        msg_body.attach(html_part)

        msg.attach(msg_body)

        for attachment in email.attachments:
            part = MIMEApplication(attachment.content)
            part.add_header("Content-Disposition", "attachment", filename=attachment.name)
            msg.attach(part)

        return dict(Source=email.sender, Destinations=email.recipients, RawMessage={"Data": msg.as_string()})



import traceback
from typing import Any, Optional
from sqlalchemy.engine import Result
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.exc import DBAPIError
from sqlalchemy.sql import Executable
from clubbi_utils.json_logging import jlogger

# https://www.postgresql.org/docs/8.2/errcodes-appendix.html
_DEADLOCK_DETECTED_ERROR_CODE = "40P01"


def is_deadlock_error(dbapie: DBAPIError) -> bool:
    return dbapie.orig.pgcode == _DEADLOCK_DETECTED_ERROR_CODE


async def retry_query_on_deadlock(
    engine: AsyncEngine,
    statement: Executable,
    parameters: Optional[Any] = None,
    maximum_number_of_retries: int = 10,
) -> Result:
    error: Exception
    for attempt in range(maximum_number_of_retries):
        async with engine.begin() as conn:
            try:
                return await conn.execute(statement, parameters)
            except DBAPIError as dbapie:
                if not is_deadlock_error(dbapie):
                    raise
                jlogger.warning(
                    "retry_query_on_deadlock",
                    "deadlock_error",
                    str_err=str(dbapie),
                    repr_err=repr(dbapie),
                    traceback=traceback.format_exc(),
                    attempt=attempt + 1,
                )
                error = dbapie
    raise error



from typing import Iterable, List, TypeVar
import itertools


T = TypeVar("T")


def iter_as_chunks(iterable: Iterable[T], n: int) -> Iterable[List[T]]:
    it = iter(iterable)
    while chunk := list(itertools.islice(it, n)):
        yield chunk



import time
from inspect import iscoroutinefunction

from clubbi_utils.logging import logger


def with_timeit(f):
    meta = dict(
        qualname=f.__qualname__,
        __module__=f.__module__,
    )

    def wrapper(*args, **kwargs):
        begin = time.time()
        r = f(*args, **kwargs)
        end = time.time()
        logger.info(dict(name=f.__name__, duration=end - begin, meta=meta))
        return r

    async def async_wrapper(*args, **kwargs):
        begin = time.time()
        r = await f(*args, **kwargs)
        end = time.time()
        logger.info(dict(name=f.__name__, duration=end - begin, meta=meta))
        return r

    if iscoroutinefunction(f):
        return async_wrapper
    else:
        return wrapper



from typing import AsyncIterator, List, Union
from clubbi_utils.logging import logger

from pathlib import Path
import tempfile
import shutil


class S3ObjectStorageLocalMock:
    def __init__(self, name: str = "S3ObjectStorageLocalMock", key_prefix: str = ""):
        self._key_prefix = key_prefix
        self._path = Path(tempfile.TemporaryDirectory().name + "-" + name, key_prefix)
        self._path.mkdir(parents=True)
        logger.info(f"initializing S3ObjectStorageLocalMock in {self.path}")

    def __del__(self):
        shutil.rmtree(self._path)

    @property
    def path(self) -> Path:
        return self._path

    async def put_object(
        self,
        name: str,
        data: Union[bytes, AsyncIterator[bytes]],
    ) -> str:
        file_path = Path(self.path, name)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if isinstance(data, bytes):
            file_path.write_bytes(data)
            return name
        with file_path.open("wb") as f:
            async for chunk in data:
                f.write(chunk)
        return name

    async def stream_object(
        self,
        key: str,
        full_key: bool = False,
        chunk_length: int = 1024 * 1024,
    ) -> AsyncIterator[bytes]:
        file_path = Path(self.path, key)
        with file_path.open("rb") as f:
            while chunk := f.read(chunk_length):
                yield chunk

    async def get_object(self, key: str, full_key: bool = False) -> bytes:
        file_path = Path(self.path, key)
        return file_path.read_bytes()

    async def list_objects(self, prefix: str = "") -> List[str]:
        dir_path = Path(self.path, prefix)
        return [str(p.relative_to(self.path)) for p in dir_path.glob("*")]

    async def copy_object(self, origin: str, destination: str) -> None:
        origin_path = Path(self.path, origin)
        destination_path = Path(self.path, destination)

        shutil.copy(origin_path, destination_path)
